package org.processmining.lifecycleperformance.algorithms;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.lifecycleperformance.models.lifecycle.LifecycleModel;
import org.processmining.lifecycleperformance.models.performance.PerformanceMeasure;
import org.processmining.lifecycleperformance.models.performance.PerformanceMetric;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;

import net.objecthunter.exp4j.Expression;
import net.objecthunter.exp4j.ExpressionBuilder;

// TODO [low] replace with Groovy code (Edu)
public class CalculatePerformanceMetricAlgorithm {

	// FIELDS

	Map<String, Double> variableValues;
	String expression;
	ExpressionBuilder builder;
	Expression exp;
	Pattern pattern;
	Matcher matcher;
	String variable;
	String measureArguments;
	String[] arguments;
	State state;
	Transition transition;

	// METHODS

	/**
	 * 
	 * Computes the performance metrics by evaluating the expression. The values
	 * for the variables that are set in the expression are obtained from the
	 * lifecycle model for which the metric is being computed.
	 * 
	 * @param metric
	 *            The performance metric
	 * @param model
	 *            The lifecycle model
	 * @return The value
	 */
	public Double compute(PerformanceMetric metric, LifecycleModel model) {
		// Keep track of all variables and their values for all measures present in the metric's expression
		variableValues = new HashMap<String, Double>();

		expression = metric.getExpression();

		// Assign all performance measure variables their values.
		for (PerformanceMeasure measure : PerformanceMeasure.values())
			match(model, expression, measure);

		// Replace the variables with their values
		for (String variable : variableValues.keySet())
			expression = expression.replace(variable, variableValues.get(variable).toString());

		/**
		 * Note that we cannot use the variable or function capability of
		 * ExpressionBuilder since Functions can only have doubles as arguments
		 * (so no state or transition names). Also, variables cannot have
		 * argument brackets. Alternatively we could remove the brackets from
		 * the performance measure texts, e.g. time_spent_min_statename instead
		 * of time_spent_min(statename), but that is more confusing.
		 */

		// Create the exp4j expression
		builder = new ExpressionBuilder(expression);
		//		if (metric.getLabel().equals("m2"))
		//			System.out.println(expression);
		// Assign the variables their values in the expression
		exp = builder.build();

		// Evaluate the expression
		return exp.evaluate();
	}

	/**
	 * Finds every occurrence of the provided measure in the expression, and
	 * replace it with the matching values stored in the lifecycle model. Note
	 * that measures are defined as functions that can have more than one
	 * argument. In case an argument (state or transition identifier) does not
	 * exist in the model, the measure cannot be retrieved and the metric cannot
	 * be calculated, and an exception is thrown.
	 * 
	 * @param model
	 *            The lifecycle model
	 * @param expression
	 *            The expression
	 * @param measure
	 *            The performance measure
	 * @return A variable->value map
	 */
	private void match(LifecycleModel model, String expression, PerformanceMeasure measure) {
		// Create a regular expression matcher for the regex that represents the measure
		pattern = Pattern.compile(measure.getRegex());
		matcher = pattern.matcher(expression);

		// For every match of this measure, replace it by the correct values.
		while (matcher.find()) {
			// Find the complete measure regex including it's arguments
			variable = matcher.group(0);

			// Extract the arguments of the measure
			measureArguments = matcher.group(1);
			arguments = measureArguments.split(",");

			// Now try to match every argument to a state or a transition in the model
			//TODO [low] use multiple arguments? not sure if necessary since values are stored in a single state or transition
			// Obtain the value stored in the model
			for (String argument : arguments) {
				double value = 0d;
				if (model.getStates().contains(argument)) {
					Set<State> states = model.getNodes();
					state = states.stream().filter(s -> s.getLabel().equals(argument)).findFirst().get();
					switch (measure) {
						case STATE_TIMES_SPENT_AVG :
							if (state.getAttributeMap()
									.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
								value = ((DescriptiveStatistics) state.getAttributeMap()
										.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).getMean();
							break;
						case STATE_TIMES_SPENT_MAX :
							if (state.getAttributeMap()
									.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
								value = ((DescriptiveStatistics) state.getAttributeMap()
										.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).getMax();
							break;
						case STATE_TIMES_SPENT_MEDIAN :
							if (state.getAttributeMap()
									.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
								value = ((DescriptiveStatistics) state.getAttributeMap()
										.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
												.getPercentile(50);
							break;
						case STATE_TIMES_SPENT_MIN :
							if (state.getAttributeMap()
									.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
								value = ((DescriptiveStatistics) state.getAttributeMap()
										.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).getMin();
							break;
						case STATE_TIMES_SPENT_SUM :
							if (state.getAttributeMap()
									.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
								value = ((DescriptiveStatistics) state.getAttributeMap()
										.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).getSum();
							break;
						case STATE_TIMES_SPENT_COUNT :
							if (state.getAttributeMap()
									.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
								value = ((DescriptiveStatistics) state.getAttributeMap()
										.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).getN();
							break;
						default :
							// Should never occur
							throw new IllegalArgumentException("Performance measure \"" + measure.getLabel()
									+ "\" undefined for state \"" + state.getLabel() + "\"");
					}
				} else if (model.getTransitions().contains(argument)) {
					Set<Transition> transitions = model.getEdges();
					// Take the first transition with this identifier (as they are assumed to be unique)
					transition = transitions.stream().filter(t -> t.getLabel().equals(argument)).findFirst().get();
					switch (measure) {
						case TRANSITION_TIMES_EXECUTED :
							if (transition.getAttributeMap()
									.containsKey(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel()))
								value = (Integer) transition.getAttributeMap()
										.get(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel());
							break;
						default :
							throw new IllegalArgumentException("Performance measure \"" + measure.getLabel()
									+ "\" undefined for transition \"" + transition.getLabel() + "\"");
					}
				}

				// Add the measure to the expression as a variable	
				variableValues.put(variable, value);
			}
		}
	}

}
