package org.processmining.lifecycleperformance.algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.lifecycleperformance.models.comparators.XEventTimeStampComparator;
import org.processmining.lifecycleperformance.models.exstensions.XArtifactLifecycleExtension;
import org.processmining.lifecycleperformance.models.lifecycle.Aggregation;
import org.processmining.lifecycleperformance.models.lifecycle.LifecycleModel;
import org.processmining.lifecycleperformance.models.lifecycle.LifecycleModelFactory;
import org.processmining.lifecycleperformance.models.performance.PerformanceAnalysisResults;
import org.processmining.lifecycleperformance.models.performance.PerformanceMeasure;
import org.processmining.lifecycleperformance.models.performance.PerformanceMetric;
import org.processmining.lifecycleperformance.parameters.LifecyclePerformanceParameters;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;

import com.google.common.collect.Table;
import com.google.common.collect.Table.Cell;

public class LifecyclePerformanceAlgorithm {

	// XES EXTENSIONS

	private XTimeExtension te = XTimeExtension.instance();
	private XArtifactLifecycleExtension le = XArtifactLifecycleExtension.instance();

	// FIELDS

	private LifecyclePerformanceParameters parameters;
	private PerformanceAnalysisResults results;

	// GETTERS AND SETTERS

	public LifecyclePerformanceParameters getParameters() {
		return parameters;
	}

	public void setParameters(LifecyclePerformanceParameters parameters) {
		this.parameters = parameters;
	}

	public PerformanceAnalysisResults getResults() {
		return results;
	}

	public void setMemory(PerformanceAnalysisResults results) {
		this.results = results;
	}

	// METHODS

	/**
	 * Initializes the memory
	 */
	protected void init() {
		results = new PerformanceAnalysisResults();
	}

	/**
	 * Measures performance measures and calculates performance metrics using
	 * the lifecycle performance approach and a static event log as input.
	 */
	protected void handleEventlog(PluginContext context, XLog eventlog) {
		// Transform event log into list of events ('stream')
		List<XEvent> eventList = new ArrayList<XEvent>();

		// Add all events in the log to the list
		for (XTrace trace : eventlog)
			eventList.addAll(trace);

		// Keep only events that affect at least one lifecycle model according to the lifecycle extension
		// Not strictly necessary, but speeds up sorting.
		eventList = eventList.stream().filter(event -> CollectionUtils.isNotEmpty(le.extractMoves(event)))
				.collect(Collectors.toList());

		// Sort the list by time (assume no events with same timestamp effect the same lifecycle model instance).
		//TODO [low] assumption for lifecycle model instance event timestamps
		Collections.sort(eventList, new XEventTimeStampComparator());

		// First, handle each event
		for (XEvent event : eventList)
			handleEvent(event);

		// Then, at the end, compute the performance metrics only once
		calculateLifecyclePerformanceMetrics();
	}

	/**
	 * Handles the given event and its potential effects on any performance
	 * measure and metric.
	 * 
	 * @param event
	 *            The event to handle
	 */
	@SuppressWarnings("unchecked")
	protected void handleEvent(XEvent event) {

		// Loop over the affected lifecycle model instances
		Table<String, String, String> table = le.extractTable(event);
		for (Cell<String, String, String> cell : table.cellSet()) {
			String model = cell.getRowKey();
			String instance = cell.getColumnKey();
			String transition = cell.getValue();

			// Get the affected lifecycle model instance from memory, and if it doesn't exist yet, create a new one.
			if (!results.getMemory().contains(model, instance))
				results.getMemory().put(model, instance,
						LifecycleModelFactory.generateLifecycleModel(model, instance, parameters));

			LifecycleModel lcm = results.getMemory().get(model, instance);

			if (!lcm.getAttributeMap().containsKey(LifecycleModel.AGGREGATION))
				lcm.getAttributeMap().put(LifecycleModel.AGGREGATION, new HashMap<String, Object>());

			for (String aggregationAttribute : parameters.getAggregations(model)) {
				if (event.getAttributes().containsKey(aggregationAttribute))
					((Map<String, Object>) lcm.getAttributeMap().get(LifecycleModel.AGGREGATION)).put(
							aggregationAttribute,
							new Aggregation(aggregationAttribute, event.getAttributes().get(aggregationAttribute)));
			}

			// Check whether the transition is possible from the current state and time. Skip if not possible (assume total conformance)
			//TODO [low] assumption for total conformance
			long eventTimeStamp = te.extractTimestamp(event).getTime();
			if (!lcm.canExecuteTransition(transition, eventTimeStamp))
				continue;

			// Find the transition (first transition with this label, assume deterministic lifecycle models)
			//TODO [low] assumption for deterministic lifecycle models
			State currentState = lcm.getCurrentState();
			Transition t = lcm.getOutEdges(currentState).stream().filter(trans -> trans.getLabel().equals(transition))
					.findFirst().get();

			// 'Execute' the transition:
			// - Add the current time spent in this state to the performance statistics of the current state
			if (!currentState.getAttributeMap().containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
				currentState.getAttributeMap().put(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel(),
						new DescriptiveStatistics());
			((DescriptiveStatistics) currentState.getAttributeMap()
					.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel()))
							.addValue(eventTimeStamp - lcm.getTimeEnteredCurrentState());

			// - Add another count to the times executed of the transition
			if (!t.getAttributeMap().containsKey(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel()))
				t.getAttributeMap().put(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel(), 0);
			t.getAttributeMap().put(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel(),
					(int) t.getAttributeMap().get(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel()) + 1);

			// - Set the new current state
			currentState = lcm.setCurrentStateIdentifier(t.getTarget().getIdentifier());

			// - Set the time that the new current state was entered
			lcm.setTimeEnteredCurrentState(eventTimeStamp);
		}
	}

	/**
	 * This method takes as input the performance measurements obtained by
	 * replaying the events on the lifecycle model instances, and calculates any
	 * performance metrics defined on them.
	 */
	protected void calculateLifecyclePerformanceMetrics() {
		// Algorithm
		CalculatePerformanceMetricAlgorithm alg = new CalculatePerformanceMetricAlgorithm();
		// For every model, for every instance, calculate the performance metrics
		for (LifecycleModel model : parameters.getLifecycleModels()) {
			Map<String, LifecycleModel> instanceModels = results.getMemory().row(model.getLabel());
			// For every model instance, calculate the performance metrics
			for (String instance : instanceModels.keySet()) {
				LifecycleModel instanceModel = instanceModels.get(instance);
				for (PerformanceMetric performanceMetric : model.getPerformanceMetrics()) {
					instanceModel.getAttributeMap().put(performanceMetric.getLabel(),
							alg.compute(performanceMetric, instanceModel));
				}
			}
		}
	}

	/**
	 * Sets the current state to the first found initial state (i.e. any state
	 * without input transitions). It is assumed that only one such state
	 * exists, but technically, that needn't be the case.
	 * 
	 * @param model
	 *            The lifecycle model
	 */
	protected void setInitialState(LifecycleModel model) {
		if (model.getCurrentStateIdentifier() != null)
			return;

		for (State state : model.getNodes())
			if (model.getInEdges(state).isEmpty())
				model.setCurrentStateIdentifier(state.getIdentifier());

		model.setTimeEnteredCurrentState(0L);
	}

}
