package org.processmining.lifecycleperformance.models.exstensions;

import java.net.URI;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeList;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;
import com.google.common.collect.Table.Cell;

/**
 * Extension that adds possibly multiple lifecycle model instance transitions to
 * events. Using this extension, events can signal a transition in multiple
 * lifecycle model instances. For example, an event can signal a transition
 * 'complete' in the case lifecycle model, instance 'case1', while at the same
 * time signaling a 'complete' transition in the activity lifecycle model of
 * instance 'activityXnr123'. Any lifecycle model can be used.
 * 
 * @author B.F.A. Hompes
 */
public class XArtifactLifecycleExtension extends XExtension {

	private static final long serialVersionUID = -4974421470391509121L;

	private XFactory factory;

	/**
	 * Properties of this XES extension. Note that this extension exists only in
	 * code, it does not have a unique URI at this moment.
	 */
	private static final String NAME;
	private static final String PREFIX;
	private static final URI URII;

	/**
	 * Singleton instance of this extension.
	 */
	private static XArtifactLifecycleExtension SINGLETON;

	/**
	 * Keys for the attribute.
	 */
	public static final String KEY_MOVES;
	public static final String KEY_MODEL;
	public static final String KEY_INSTANCE;
	public static final String KEY_TRANSITION;

	/**
	 * Attribute prototype.
	 */
	public static XAttributeList ATTR_MOVES;

	static {
		NAME = "Artifact Lifecycle";
		PREFIX = "artifactlifecycle:";
		URII = URI.create("none");

		KEY_MOVES = "moves";
		KEY_MODEL = "model";
		KEY_INSTANCE = "instance";
		KEY_TRANSITION = "transition";

		SINGLETON = new XArtifactLifecycleExtension();
	}

	/**
	 * Private constructor
	 */
	private XArtifactLifecycleExtension() {
		super(NAME, PREFIX, URII);
		factory = XFactoryRegistry.instance().currentDefault();
		ATTR_MOVES = factory.createAttributeList(PREFIX + KEY_MOVES, this);
		this.eventAttributes.add((XAttributeList) ATTR_MOVES.clone());
	}

	/**
	 * Provides static access to the singleton instance of this extension.
	 * 
	 * @return Singleton instance.
	 */
	public static XArtifactLifecycleExtension instance() {
		return SINGLETON;
	}

	private Object readResolve() {
		return SINGLETON;
	}

	public Collection<XAttribute> extractMoves(XEvent event) {
		XAttribute moves = event.getAttributes().get(PREFIX + KEY_MOVES);
		if (moves == null) {
			return new HashSet<XAttribute>();
		} else {
			return ((XAttributeList) moves).getCollection();
		}
	}

	public Table<String, String, String> extractTable(XEvent event) {
		XAttribute moves = event.getAttributes().get(PREFIX + KEY_MOVES);
		if (moves == null) {
			return null;
		} else {
			Table<String, String, String> table = HashBasedTable.create();
			Collection<XAttribute> collection = ((XAttributeList) moves).getCollection();
			// Loop over all the lifecycle entries in the list
			Iterator<XAttribute> it = collection.iterator();
			while (it.hasNext()) {
				XAttributeLiteral model = (XAttributeLiteral) it.next();
				/*
				 * Only consider lifecycle model attributes (no other attributes
				 * should exist...) that have meta-attributes for instance and
				 * transition.
				 */
				if (!model.getKey().equals(PREFIX + KEY_MODEL) || !model.hasAttributes())
					continue;
				// Get the model name
				String valueModel = model.getValue();
				// Get the instance meta-attribute
				XAttributeLiteral instance = (XAttributeLiteral) model.getAttributes().get(PREFIX + KEY_INSTANCE);
				if (instance == null)
					continue;
				// Get the transition name
				String valueInstance = instance.getValue();
				// Get the transition meta-attribute
				XAttributeLiteral transition = (XAttributeLiteral) model.getAttributes().get(PREFIX + KEY_TRANSITION);
				if (transition == null)
					continue;
				// Get the transition name
				String valueTransition = transition.getValue();
				// Save the model, instance, transition entry.
				table.put(valueModel, valueInstance, valueTransition);
			}
			return table;
		}
	}

	public void assignMoves(XEvent event, Collection<XAttribute> collection) {
		if (collection != null && collection.size() > 0) {
			XAttributeList moves = (XAttributeList) ATTR_MOVES.clone();

			for (XAttribute model : collection) {
				// Only add the model attribute if it has meta-attributes for instance and transition.
				if (model.getAttributes().containsKey(PREFIX + KEY_INSTANCE)
						&& model.getAttributes().containsKey(PREFIX + KEY_TRANSITION))
					moves.addToCollection(model);
			}

			event.getAttributes().put(PREFIX + KEY_MOVES, moves);
		}
	}

	public void assignTable(XEvent event, Table<String, String, String> table) {
		XAttributeList moves = (XAttributeList) ATTR_MOVES.clone();

		for (Cell<String, String, String> cell : table.cellSet()) {
			XAttributeLiteral model = factory.createAttributeLiteral(PREFIX + KEY_MODEL, cell.getRowKey(), this);
			XAttributeLiteral instance = factory.createAttributeLiteral(PREFIX + KEY_INSTANCE, cell.getColumnKey(),
					this);
			XAttributeLiteral transition = factory.createAttributeLiteral(PREFIX + KEY_TRANSITION, cell.getValue(),
					this);
			model.getAttributes().put(PREFIX + KEY_INSTANCE, instance);
			model.getAttributes().put(PREFIX + KEY_TRANSITION, transition);
			moves.addToCollection(model);
		}

		event.getAttributes().put(PREFIX + KEY_MOVES, moves);
	}

	public void add(XEvent event, String model, String instance, String transition) {
		if (extractMoves(event) == null || extractMoves(event).isEmpty())
			event.getAttributes().put(PREFIX + KEY_MOVES, (XAttributeList) ATTR_MOVES.clone());

		XAttributeLiteral attrModel = factory.createAttributeLiteral(PREFIX + KEY_MODEL, model, this);
		XAttributeLiteral attrInstance = factory.createAttributeLiteral(PREFIX + KEY_INSTANCE, instance, this);
		XAttributeLiteral attrTransition = factory.createAttributeLiteral(PREFIX + KEY_TRANSITION, transition, this);

		attrModel.getAttributes().put(PREFIX + KEY_INSTANCE, attrInstance);
		attrModel.getAttributes().put(PREFIX + KEY_TRANSITION, attrTransition);

		extractMoves(event).add(attrModel);
	}

}
