package org.processmining.lifecycleperformance.models.lifecycle;

import java.util.NoSuchElementException;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.lifecycleperformance.models.performance.PerformanceMeasure;
import org.processmining.lifecycleperformance.models.performance.PerformanceMetric;
import org.processmining.lifecycleperformance.parameters.LifecyclePerformanceParameters;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;

public class LifecycleModelFactory {

	public static LifecycleModel generateLifecycleModel(String label, String instance,
			LifecyclePerformanceParameters parameters) {
		try {

			// Find the lifecycle model in the parameters, if it exists, else return a new empty model
			LifecycleModel paramModel = parameters.getLifecycleModels().stream().filter(m -> m.getLabel().equals(label))
					.findFirst().get();

			// Create an new instance of this model
			LifecycleModel model = new LifecycleModel(paramModel, instance);
			// Set the model to be in its initial state.
			model.setCurrentStateIdentifier(paramModel.getCurrentStateIdentifier());
			model.setTimeEnteredCurrentState(0l);

			return model;

		} catch (NoSuchElementException ex) {

			// Create a new empty model
			System.err.println("Could not find lifecycle model '" + label + "' for instance '" + instance
					+ "', creating an empty lifecycle model instead.");
			return new LifecycleModel(label, instance);

		}

	}

	public static LifecycleModel generateStandardActivityLifecycleModel() {
		LifecycleModel standard = new LifecycleModel("standard");

		standard.addState("init");
		standard.addState("scheduled");
		standard.addState("assigned");
		standard.addState("started");
		standard.addState("suspended");
		standard.addState("completed successfully");
		standard.addState("completed unsuccessfully");

		standard.addTransition("init", "scheduled", "schedule");
		standard.addTransition("scheduled", "assigned", "assign");
		standard.addTransition("assigned", "assigned", "reassign");
		standard.addTransition("assigned", "started", "start");
		standard.addTransition("started", "suspended", "suspend");
		standard.addTransition("suspended", "started", "resume");

		standard.addTransition("init", "completed successfully", "autoskip");
		standard.addTransition("scheduled", "completed successfully", "manualskip");
		standard.addTransition("assigned", "completed successfully", "manualskip");
		standard.addTransition("started", "completed successfully", "complete");

		standard.addTransition("scheduled", "completed unsuccessfully", "withdraw");
		standard.addTransition("assigned", "completed unsuccessfully", "withdraw");
		standard.addTransition("started", "completed unsuccessfully", "abort_activity");
		standard.addTransition("init", "completed unsuccessfully", "abort_case");
		standard.addTransition("scheduled", "completed unsuccessfully", "abort_case");
		standard.addTransition("assigned", "completed unsuccessfully", "abort_case");
		standard.addTransition("started", "completed unsuccessfully", "abort_case");
		standard.addTransition("suspended", "completed unsuccessfully", "abort_case");

		standard.setCurrentStateIdentifier("init");
		standard.setTimeEnteredCurrentState(0l);

		PerformanceMetric duration = new PerformanceMetric("duration");
		PerformanceMetric waitingTime = new PerformanceMetric("waitingtime");
		PerformanceMetric sojournTime = new PerformanceMetric("sojourntime");
		PerformanceMetric suspensions = new PerformanceMetric("suspensions");
		PerformanceMetric reassignments = new PerformanceMetric("reassignments");

		duration.setExpression("sum(times_spent(started))");
		waitingTime.setExpression("sum(times_spent(scheduled))");
		sojournTime.setExpression(
				"sum(times_spent(scheduled))+sum(times_spent(assigned))+sum(times_spent(started))+sum(times_spent(suspended))");
		suspensions.setExpression("count(times_spent(suspended))");
		reassignments.setExpression("times_executed(reassign)");

		standard.getPerformanceMetrics().add(duration);
		standard.getPerformanceMetrics().add(waitingTime);
		standard.getPerformanceMetrics().add(sojournTime);
		standard.getPerformanceMetrics().add(suspensions);
		standard.getPerformanceMetrics().add(reassignments);

		return standard;
	}

	public static LifecycleModel generateSmallActivityLifecycleModel() {
		LifecycleModel standard = new LifecycleModel("standard");

		standard.addState("init");
		standard.addState("started");
		standard.addState("completed");

		standard.addTransition("init", "started", "start");
		standard.addTransition("started", "completed", "complete");

		standard.setCurrentStateIdentifier("init");
		standard.setTimeEnteredCurrentState(0l);

		PerformanceMetric duration = new PerformanceMetric("duration");
		duration.setExpression("sum(times_spent(started))");
		standard.getPerformanceMetrics().add(duration);

		return standard;
	}

	public static LifecycleModel generateBPAFActivityLifecycleModel() {
		return new LifecycleModel("BPAF");
	}

	public static LifecycleModel aggregateModels(LifecycleModel... models) {
		/**
		 * Create a copy from the first model in the collection of models to be
		 * aggregated, as the selected models should have the same structure and
		 * metrics defined on them.
		 */
		LifecycleModel aggregateModel = new LifecycleModel(models[0]);

		// Copy the measures from the instance level to the aggregate level
		for (LifecycleModel instanceModel : models) {
			for (State instanceState : instanceModel.getNodes()) {
				if (instanceState.getAttributeMap()
						.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())) {
					State modelState = aggregateModel.getNode(instanceState.getIdentifier());
					if (!modelState.getAttributeMap()
							.containsKey(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())) {
						modelState.getAttributeMap().put(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel(),
								new DescriptiveStatistics());
					}
					for (double value : ((DescriptiveStatistics) instanceState.getAttributeMap()
							.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).getValues()) {
						((DescriptiveStatistics) modelState.getAttributeMap()
								.get(PerformanceMeasure.STATE_TIMES_SPENT_SUMMARY.getLabel())).addValue(value);
					}
				}
			}
			for (Transition instanceTransition : instanceModel.getEdges()) {
				if (instanceTransition.getAttributeMap()
						.containsKey(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel())) {
					Transition modelTransition = aggregateModel.getEdges(instanceTransition.getIdentifier()).iterator()
							.next();
					if (!modelTransition.getAttributeMap()
							.containsKey(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel())) {
						modelTransition.getAttributeMap().put(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel(),
								0);
					}
					modelTransition.getAttributeMap().put(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel(),
							(int) modelTransition.getAttributeMap()
									.get(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel())
									+ (int) instanceTransition.getAttributeMap()
											.get(PerformanceMeasure.TRANSITION_TIMES_EXECUTED.getLabel()));
				}
			}
		}

		for (PerformanceMetric performanceMetric : aggregateModel.getPerformanceMetrics()) {
			DescriptiveStatistics stats = new DescriptiveStatistics();
			for (LifecycleModel instanceModel : models) {
				stats.addValue((double) instanceModel.getAttributeMap().get(performanceMetric.getLabel()));
			}
			aggregateModel.getAttributeMap().put(performanceMetric.getLabel(), stats);
		}

		return aggregateModel;
	}

}
